@@ -22,12 +22,11 @@
 #include "vmcstream.h"
 #include "spice-channel-priv.h"
 #include "gio-coroutine.h"
-#include "glib-compat.h"
 
 struct _SpiceVmcInputStream
 {
     GInputStream parent_instance;
-    GSimpleAsyncResult *result;
+    GTask *task;
     struct coroutine *coroutine;
 
     SpiceChannel *channel;
@@ -36,7 +35,6 @@ struct _SpiceVmcInputStream
     gsize count;
     gsize pos;
 
-    GCancellable *cancellable;
     gulong cancel_id;
 };
 
@@ -99,8 +97,26 @@ spice_vmc_input_stream_new(void)
     return self;
 }
 
+typedef struct _complete_in_idle_cb_data {
+    GTask *task;
+    gssize pos;
+} complete_in_idle_cb_data;
+
+static gboolean
+complete_in_idle_cb(gpointer user_data)
+{
+    complete_in_idle_cb_data *data = user_data;
+
+    g_task_return_int(data->task, data->pos);
+
+    g_object_unref (data->task);
+    g_free (data);
+
+    return FALSE;
+}
+
 /* coroutine */
-/**
+/*
  * Feed a SpiceVmc stream with new data from a coroutine
  *
  * The other end will be waiting on read_async() until data is fed
@@ -118,11 +134,13 @@ spice_vmc_input_stream_co_data(SpiceVmcInputStream *self,
     self->coroutine = coroutine_self();
 
     while (size > 0) {
-        SPICE_DEBUG("spicevmc co_data %p", self->result);
-        if (!self->result)
+        complete_in_idle_cb_data *cb_data;
+
+        SPICE_DEBUG("spicevmc co_data %p", self->task);
+        if (!self->task)
             coroutine_yield(NULL);
 
-        g_return_if_fail(self->result != NULL);
+        g_return_if_fail(self->task != NULL);
 
         gsize min = MIN(self->count, size);
         memcpy(self->buffer, data, min);
@@ -139,14 +157,15 @@ spice_vmc_input_stream_co_data(SpiceVmcInputStream *self,
         if (self->all && min > 0 && self->pos != self->count)
             continue;
 
-        g_simple_async_result_set_op_res_gssize(self->result, self->pos);
+        /* Let's deal with the task complete in idle by ourselves, as GTask
+         * heuristic only makes sense in a non-coroutine case.
+         */
+        cb_data = g_new(complete_in_idle_cb_data , 1);
+        cb_data->task = g_object_ref(self->task);
+        cb_data->pos = self->pos;
+        g_idle_add(complete_in_idle_cb, cb_data);
 
-        g_simple_async_result_complete_in_idle(self->result);
-        g_clear_object(&self->result);
-        if (self->cancellable) {
-            g_cancellable_disconnect(self->cancellable, self->cancel_id);
-            g_clear_object(&self->cancellable);
-        }
+        g_clear_object(&self->task);
     }
 
     self->coroutine = NULL;
@@ -158,19 +177,15 @@ read_cancelled(GCancellable *cancellable,
 {
     SpiceVmcInputStream *self = SPICE_VMC_INPUT_STREAM(user_data);
 
-    SPICE_DEBUG("read cancelled, %p", self->result);
-    g_simple_async_result_set_error(self->result,
-                                    G_IO_ERROR, G_IO_ERROR_CANCELLED,
-                                    "read cancelled");
-    g_simple_async_result_complete_in_idle(self->result);
+    SPICE_DEBUG("read cancelled, %p", self->task);
+    g_task_return_new_error(self->task,
+                            G_IO_ERROR, G_IO_ERROR_CANCELLED,
+                            "read cancelled");
 
-    g_clear_object(&self->result);
-
-    /* See FIXME */
-    /* if (self->cancellable) { */
-    /*     g_cancellable_disconnect(self->cancellable, self->cancel_id); */
-    /*     g_clear_object(&self->cancellable); */
-    /* } */
+    /* With GTask, we don't need to disconnect GCancellable when task is
+     * cancelled within cancellable callback as it could lead to deadlocks
+     * e.g: https://bugzilla.gnome.org/show_bug.cgi?id=705395 */
+    g_clear_object(&self->task);
 }
 
 G_GNUC_INTERNAL void
@@ -183,21 +198,19 @@ spice_vmc_input_stream_read_all_async(GInputStream        *stream,
                                       gpointer             user_data)
 {
     SpiceVmcInputStream *self = SPICE_VMC_INPUT_STREAM(stream);
-    GSimpleAsyncResult *result;
+    GTask *task;
 
     /* no concurrent read permitted by ginputstream */
-    g_return_if_fail(self->result == NULL);
-    g_return_if_fail(self->cancellable == NULL);
+    g_return_if_fail(self->task == NULL);
     self->all = TRUE;
     self->buffer = buffer;
     self->count = count;
     self->pos = 0;
-    result = g_simple_async_result_new(G_OBJECT(self),
-                                       callback,
-                                       user_data,
-                                       spice_vmc_input_stream_read_async);
-    self->result = result;
-    self->cancellable = g_object_ref(cancellable);
+    task = g_task_new(self,
+                      cancellable,
+                      callback,
+                      user_data);
+    self->task = task;
     if (cancellable)
         self->cancel_id =
             g_cancellable_connect(cancellable, G_CALLBACK(read_cancelled), self, NULL);
@@ -211,27 +224,17 @@ spice_vmc_input_stream_read_all_finish(GInputStream *stream,
                                        GAsyncResult *result,
                                        GError **error)
 {
-    GSimpleAsyncResult *simple;
+    GTask *task = G_TASK(result);
     SpiceVmcInputStream *self = SPICE_VMC_INPUT_STREAM(stream);
+    GCancellable *cancel;
 
-    g_return_val_if_fail(g_simple_async_result_is_valid(result,
-                                                        G_OBJECT(self),
-                                                        spice_vmc_input_stream_read_async),
-                         -1);
-
-    simple = (GSimpleAsyncResult *)result;
-
-    /* FIXME: calling _finish() is required. Disconnecting in
-       read_cancelled() causes a deadlock. #705395 */
-    if (self->cancellable) {
-        g_cancellable_disconnect(self->cancellable, self->cancel_id);
-        g_clear_object(&self->cancellable);
+    g_return_val_if_fail(g_task_is_valid(task, self), -1);
+    cancel = g_task_get_cancellable(task);
+    if (!g_cancellable_is_cancelled(cancel)) {
+         g_cancellable_disconnect(cancel, self->cancel_id);
+         self->cancel_id = 0;
     }
-
-    if (g_simple_async_result_propagate_error(simple, error))
-        return -1;
-
-    return g_simple_async_result_get_op_res_gssize(simple);
+    return g_task_propagate_int(task, error);
 }
 
 static void
@@ -244,21 +247,17 @@ spice_vmc_input_stream_read_async(GInputStream        *stream,
                                   gpointer             user_data)
 {
     SpiceVmcInputStream *self = SPICE_VMC_INPUT_STREAM(stream);
-    GSimpleAsyncResult *result;
+    GTask *task;
 
     /* no concurrent read permitted by ginputstream */
-    g_return_if_fail(self->result == NULL);
-    g_return_if_fail(self->cancellable == NULL);
+    g_return_if_fail(self->task == NULL);
     self->all = FALSE;
     self->buffer = buffer;
     self->count = count;
     self->pos = 0;
-    result = g_simple_async_result_new(G_OBJECT(self),
-                                       callback,
-                                       user_data,
-                                       spice_vmc_input_stream_read_async);
-    self->result = result;
-    self->cancellable = g_object_ref(cancellable);
+
+    task = g_task_new(self, cancellable, callback, user_data);
+    self->task = task;
     if (cancellable)
         self->cancel_id =
             g_cancellable_connect(cancellable, G_CALLBACK(read_cancelled), self, NULL);
@@ -272,27 +271,18 @@ spice_vmc_input_stream_read_finish(GInputStream *stream,
                                    GAsyncResult *result,
                                    GError **error)
 {
-    GSimpleAsyncResult *simple;
+    GTask *task = G_TASK(result);
     SpiceVmcInputStream *self = SPICE_VMC_INPUT_STREAM(stream);
+    GCancellable *cancel;
 
-    g_return_val_if_fail(g_simple_async_result_is_valid(result,
-                                                        G_OBJECT(self),
-                                                        spice_vmc_input_stream_read_async),
-                         -1);
+    g_return_val_if_fail(g_task_is_valid(task, self), -1);
 
-    simple = (GSimpleAsyncResult *)result;
-
-    /* FIXME: calling _finish() is required. Disconnecting in
-       read_cancelled() causes a deadlock. #705395 */
-    if (self->cancellable) {
-        g_cancellable_disconnect(self->cancellable, self->cancel_id);
-        g_clear_object(&self->cancellable);
+    cancel = g_task_get_cancellable(task);
+    if (!g_cancellable_is_cancelled(cancel)) {
+         g_cancellable_disconnect(cancel, self->cancel_id);
+         self->cancel_id = 0;
     }
-
-    if (g_simple_async_result_propagate_error(simple, error))
-        return -1;
-
-    return g_simple_async_result_get_op_res_gssize(simple);
+    return g_task_propagate_int(task, error);
 }
 
 static gssize
@@ -407,11 +397,14 @@ spice_vmc_output_stream_write_finish(GOutputStream *stream,
                                      GError **error)
 {
     SpiceVmcOutputStream *self = SPICE_VMC_OUTPUT_STREAM(stream);
-    GSimpleAsyncResult *res =
-        g_simple_async_result_get_op_res_gpointer(G_SIMPLE_ASYNC_RESULT(simple));
+    GAsyncResult *res = g_task_propagate_pointer(G_TASK(simple), error);
+    gssize bytes_written;
 
     SPICE_DEBUG("spicevmc write finish");
-    return spice_vmc_write_finish(self->channel, G_ASYNC_RESULT(res), error);
+    bytes_written = spice_vmc_write_finish(self->channel, res, error);
+    g_object_unref(res);
+
+    return bytes_written;
 }
 
 static void
@@ -419,12 +412,11 @@ write_cb(GObject *source_object,
          GAsyncResult *res,
          gpointer user_data)
 {
-    GSimpleAsyncResult *simple = user_data;
+    GTask *task = user_data;
 
-    g_simple_async_result_set_op_res_gpointer(simple, res, NULL);
+    g_task_return_pointer(task, g_object_ref(res), g_object_unref);
 
-    g_simple_async_result_complete(simple);
-    g_object_unref(simple);
+    g_object_unref(task);
 }
 
 static void
@@ -437,16 +429,15 @@ spice_vmc_output_stream_write_async(GOutputStream *stream,
                                     gpointer user_data)
 {
     SpiceVmcOutputStream *self = SPICE_VMC_OUTPUT_STREAM(stream);
-    GSimpleAsyncResult *simple;
+    GTask *task;
 
     SPICE_DEBUG("spicevmc write async");
     /* an AsyncResult to forward async op to channel */
-    simple = g_simple_async_result_new(G_OBJECT(self), callback, user_data,
-                                       spice_vmc_output_stream_write_async);
+    task = g_task_new(self, cancellable, callback, user_data);
 
     spice_vmc_write_async(self->channel, buffer, count,
                           cancellable, write_cb,
-                          simple);
+                          task);
 }
 
 /* STREAM */
